program main
  use iso_c_binding
  implicit none
  include 'mpif.h'
  include 'tree.f90'
  type(c_ptr) :: tree_ctx
  real*8 :: pt_coord(100000*3)
  integer*4 :: Npt, mpi_rank, mpi_size, i, ierr
  real*4 :: rand

  integer*4 :: Ndata, Ncnt
  real*8, pointer :: pt_data(:)
  integer*4, pointer :: pt_cnt(:)
  type(c_ptr) :: pt_data_, pt_cnt_

  integer*4 :: Nnodes
  real*4, pointer :: node_coord(:)
  integer*4, pointer :: parent_lst(:), child_lst(:)
  integer*1, pointer :: node_depth(:), node_ghost(:), node_leaf(:)
  type(c_ptr) :: node_coord_, parent_lst_, child_lst_, node_depth_, node_ghost_, node_leaf_

  real*8, allocatable :: mult_coeff(:)
  integer*4, allocatable :: mult_cnt(:)


  call MPI_INIT(ierr)
  call MPI_COMM_RANK(MPI_COMM_WORLD, mpi_rank, ierr) 
  call MPI_COMM_SIZE(MPI_COMM_WORLD, mpi_size, ierr) 
  call srand(mpi_rank)


  ! create random points
  Npt = 100000
  call create_points(pt_coord, Npt)



  ! Create tree and refine using pt_coord with max 100 points per box
  call CreateTree(tree_ctx)
  call UpdateRefinement(pt_coord, Npt, 100, .true., .false., tree_ctx)



  ! Read tree node structure
  call GetTree(node_coord_, parent_lst_, child_lst_, node_depth_, node_ghost_, node_leaf_, Nnodes, tree_ctx)
  call c_f_pointer(node_coord_, node_coord, [Nnodes*3])
  call c_f_pointer(parent_lst_, parent_lst, [Nnodes])
  call c_f_pointer(child_lst_ , child_lst , [Nnodes*8])
  call c_f_pointer(node_depth_, node_depth, [Nnodes])
  call c_f_pointer(node_ghost_, node_ghost, [Nnodes])
  call c_f_pointer(node_leaf_ , node_leaf , [Nnodes])



  ! Add particles to the tree and associate some data with each particle
  call AddParticles('src_pts', pt_coord, Npt, tree_ctx)
  call AddParticleData('src_coord', 'src_pts', pt_coord, Npt*3, tree_ctx)

  ! Read the data and verify it is the same as original array
  call GetParticleData(pt_data_, Ndata, 'src_coord', tree_ctx)
  call c_f_pointer(pt_data_, pt_data, [Ndata])
  call compare_arrays(pt_coord, Npt*3, pt_data, Ndata)



  ! Read node-wise particle data
  call GetData(pt_data_, Ndata, pt_cnt_, Ncnt, 'src_coord', tree_ctx)
  call c_f_pointer(pt_data_, pt_data, [Ndata])
  call c_f_pointer(pt_cnt_, pt_cnt, [Ncnt])



  ! Write tree node structure to VTK file
  call WriteTreeVTK('tree', .false., tree_ctx)

  ! Write points and associated data to VTK file
  call WriteParticleVTK('pts', 'src_coord', .false., tree_ctx)

  ! Delete the particles (or the associated data)
  call DeleteParticleData('src_pts', tree_ctx)




  allocate(mult_coeff(Nnodes*10))
  allocate(mult_cnt(Nnodes))
  do i = 1, Nnodes
    mult_cnt(i) = 1
  enddo
  call AddData('mult_coeff', mult_coeff, Nnodes*10, mult_cnt, Nnodes, tree_ctx)
  call DeleteData('mult_coeff', tree_ctx)



  ! delete the tree
  call DeleteTree(tree_ctx)

  call MPI_FINALIZE(ierr)
end

subroutine create_points(coord, Npt)
  implicit none
  integer*4 :: Npt, i
  real*8 :: coord(3*Npt), phi, theta
  real*4 :: rand

  do i = 0, Npt-1
    phi = 2 * 3.1415926 * rand(0);
    theta = 3.1415926 * rand(0);
    coord(i*3+1) = 0.5 + 0.1125 * sin(theta) * cos(phi)
    coord(i*3+2) = 0.5 + 0.1125 * sin(theta) * sin(phi)
    coord(i*3+3) = 0.5 + 0.45 * cos(theta)
  enddo

endsubroutine

subroutine compare_arrays(A, NA, B, NB)
  implicit none
  integer*4 :: i, NA, NB
  real*8 :: A(NA), B(NB)

  if (NA /= NB) then
    print *, 'Array sizes dont match'
    call exit(1)
  endif
  do i = 1, NA
    if (A(i) /= B(i)) then
      print *, 'Array values dont match'
      call exit(1)
    endif
  enddo
endsubroutine

